# Official implementation of the Distributed Video Transcoding Protocol (DVTP) server
# Copyright (C) 2023 Dymik739
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Email: user@109.86.70.81

import os
import sys
import time
import math
import json
import socket
import hashlib
import threading

CHUNK_CHECK_INTERVAL = 60
RECOMMENDED_PING_INTERVAL = 30
CHUNK_PING_AWAIT_SECONDS = 300
SERVER_SOCKET_TIMEOUT = 120
CONNECTION_TIMEOUT_DURING_LOCK = 10

next_worker_id = 0

class Worker:
    def __init__(self, worker_id, endpoint):
        self.endpoint = endpoint
        self.last_ping_timestamp = time.time()
        self.id = worker_id

    def __str__(self):
        return f"Worker(id = {self.worker_id}, endpoint = {self.endpoint}, last_ping_timestamp = {self.last_ping_timestamp}, is_alive = {self.is_alive()}"

    def update_timestamp(self):
        self.last_seen_timestamp = time.time()

    def is_alive(self):
        return time.time() - self.last_ping_timestamp <= CHUNK_PING_AWAIT_SECONDS

class Chunk:
    def __init__(self, base_file, seq_id, start, length):
        self.base_file = base_file
        self.seq_id = seq_id
        self.start = start
        self.length = length
        self.assigned_workers = []
        self.done = False

    def __str__(self):
        return f"Chunk(seq_id = {self.seq_id}, start = {self.start}, length = {self.length}, assigned_workers = [" + \
                ', '.join([str(worker) for worker in self.assigned_workers]) + f"], done = {self.done})"

    def status(self):
        if self.done:
            return "done"
        elif not any([i.is_alive() for i in self.assigned_workers]):
            return "wanted"
        else:
            return "taken"

    def add_worker(self, worker):
        self.assigned_workers.append(worker)

    def remove_worker(self, worker_id):
        self.assigned_workers = [i for i in self.assigned_workers if i.id != worker_id]

    def get_worker_by_id(self, worker_id):
        found_workers = [i for i in self.assigned_workers if i.id == worker_id]

        if found_workers:
            return found_workers[0]
        else:
            return None
        

class ChunkQueue:
    def __init__(self):
        self.queue = []

    def __str__(self):
        return "ChunkQueue(" + ', '.join([str(chunk) for chunk in self.queue]) + ")"

    def push(self, chunk):
        self.queue.append(chunk)

    def next_wanted(self):
        available_chunks = list(filter(lambda x: x.status() == "wanted", self.queue))

        if len(available_chunks) > 0:
            return available_chunks[0]
        else:
            return None

    def get_chunk_by_id(self, seq_id):
        found_chunks = [i for i in self.queue if i.seq_id == seq_id]

        if len(found_chunks) > 0:
            return found_chunks[0]
        else:
            return None

    def processing_finished(self):
        remaining_chunks = [i for i in self.queue if not i.done]

        if len(remaining_chunks):
            return False
        else:
            return True

print("[init:fs] Creating required directories and files...")

for i in ["chunks/"]:
    if not os.path.exists(i):
        print("[init:fs] Creating {i}")
        os.mkdir(i)

if not os.path.exists("ready.txt"):
    print("[init:fs] Creating ready.txt")
    os.system("touch ready.txt")


print("[init:chunk_queue] Initializing ChunkQueue...")
chunk_queue = ChunkQueue()

print("[init:args] Parsing CLI arguments...")

fn = sys.argv[1]
duration = float(sys.argv[2])
chunk_length = int(sys.argv[3])
server_bind_addr = sys.argv[4].split(":")
server_bind_addr[1] = int(server_bind_addr[1])
script_url = sys.argv[5]
video_url = sys.argv[6]

print("[init:video] Reading file size...")
file_size = os.stat(fn).st_size
print(f"[init:video] File size: {file_size} bytes (~{round(file_size/1024/1024, 2)} MB)")

print("[init:video] Computing checksum... (might take some time)")
with open(fn, 'rb') as f:
    file_hash = hashlib.sha256(f.read()).hexdigest()
#file_hash = hashlib.file_digest(open(fn, 'rb'), 'sha256').hexdigest()
print(f"[init:video] Checksum for the file: {file_hash}")

total_chunk_amount = math.ceil(duration / chunk_length)
print(f"[init:video] Splitting video into {total_chunk_amount} chunks ({chunk_length}s each)")

present_chunks = [int(i) for i in open("ready.txt", "r").read().split("\n") if i != '']
required_chunks = [i for i in range(total_chunk_amount) if i not in present_chunks]

print(f"[init:video] {len(present_chunks)}/{total_chunk_amount} chunks are ready, {len(required_chunks)} left")

for c in required_chunks:
    chunk_queue.push(Chunk(fn, c, c*chunk_length, chunk_length))

print(f"[init:video] Added {len(chunk_queue.queue)} chunks to the queue, starting the server...")


main_server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
main_server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
main_server_socket.bind(tuple(server_bind_addr))
main_server_socket.listen()

lock = threading.Lock()
client_threads = []

def thread_cleanup_job():
    global lock, client_threads
    while True:
        time.sleep(300)
        print("[cleanup] Removing finished threads...")
        with lock:
            worker_amount_before = len(client_threads)
            client_threads = [i for i in client_threads if i.is_alive()]
            worker_amount_after = len(client_threads)
        
        removed_amount = worker_amount_before - worker_amount_after
        print(f"[cleanup] Worker threads {worker_amount_before} -> {worker_amount_after} (cleared {removed_amount} threads)")



def connection_processor(conn, addr):
    global lock, next_worker_id
    global video_url, script_url, file_size, file_hash

    print(f"[{addr[0]}:{addr[1]} - init] New connection opened")

    conn.settimeout(SERVER_SOCKET_TIMEOUT)

    databuffer = bytes()
    connection_alive = True

    receiving_binary_data = 0

    while connection_alive:
        while (bytes("\n", "UTF-8") not in databuffer) and connection_alive:
            new_data = conn.recv(1024)
            if len(new_data) == 0:
                connection_alive = False
                break

            databuffer += new_data

        if not connection_alive:
            conn.close()
            connection_alive = False
            break

        next_line = databuffer.split(bytes("\n", "UTF-8"))[0]
        databuffer = databuffer[len(next_line)+1:]
        #print(f"[network:debug] Updated databuffer: {databuffer}")
        header = next_line.decode("UTF-8").split(" ")
        
        print(f"[{addr[0]}:{addr[1]} - parser] Received command: {' '.join(header)}")

        if header[0] == "DEBUG":
            with lock:
                conn.send(f"DEBUG\nCHUNKS {str(chunk_queue)}\n".encode("UTF-8"))
        elif header[0] == "HASH":
            conn.send(f"HASH {file_hash}\n".encode("UTF-8"))
        elif header[0] == "SIZE":
            conn.send(f"SIZE {file_size}\n".encode("UTF-8"))
        elif header[0] == "URL":
            conn.send(f"URL {video_url}\n".encode("UTF-8"))
        elif header[0] == "SCRIPT":
            conn.send(f"SCRIPT {script_url}\n".encode("UTF-8"))
        elif header[0] == "NEXT":
            with lock:
                next_chunk = chunk_queue.next_wanted()
                if next_chunk:
                    new_worker = Worker(next_worker_id, addr)
                    next_worker_id += 1

                    next_chunk.add_worker(new_worker)

                    conn.send(f"CHUNK {file_hash} {next_chunk.seq_id} {next_chunk.start} {next_chunk.length} {RECOMMENDED_PING_INTERVAL} {new_worker.id}\n".encode("UTF-8"))
                else:
                    if chunk_queue.processing_finished():
                        conn.send("FINISH\n".encode("UTF-8"))
                    else:
                        conn.send(f"WAIT {CHUNK_CHECK_INTERVAL}\n".encode("UTF-8"))
        elif header[0] == "PING":
            if file_hash != header[1]:
                print(f"[{addr[0]}:{addr[1]} - PING] Hash mismatch: expected {file_hash}, got {header[1]}")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break
            
            if not header[2].isdigit():
                print(f"[{addr[0]}:{addr[1]} - PING] invalid seq_id: not an int (got {header[2]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[3].isdigit():
                print(f"[{addr[0]}:{addr[1]} - PING] invalid worker_id: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            with lock:
                chunk = chunk_queue.get_chunk_by_id(int(header[2]))

                if not chunk:
                    print(f"[{addr[0]}:{addr[1]} - PING] invalid seq_id: no chunk with id {header[2]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                worker = chunk.get_worker_by_id(int(header[3]))

                if not worker:
                    print(f"[{addr[0]}:{addr[1]} - PING] invalid worker_id: no worker with id {header[3]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                if worker.endpoint != addr:
                    print(f"[{addr[0]}:{addr[1]} - PING] Warning: worker {worker.id} for chunk {chunk.seq_id} changed it's endpoint ({worker.endpoint[0]}:{worker.endpoint[1]} -> {addr[0]}:{addr[1]})")
                    worker.endpoint = addr

                worker.update_timestamp()

                if chunk.done:
                    status = "done"
                else:
                    status = "waiting"

                conn.send(f"PONG {status}\n".encode("UTF-8"))
        elif header[0] == "UPLOAD":
            if file_hash != header[1]:
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] Hash mismatch: expected {file_hash}, got {header[1]}")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break
            
            if not header[2].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid seq_id: not an int (got {header[2]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[3].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid worker_id: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[4].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid chunk_size: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            with lock:
                chunk = chunk_queue.get_chunk_by_id(int(header[2]))

                if not chunk:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid seq_id: no chunk with id {header[2]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                worker = chunk.get_worker_by_id(int(header[3]))

                if not worker:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid worker_id: no worker with id {header[3]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                if worker.endpoint != addr:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] Warning: worker {worker.id} for chunk {chunk.seq_id} changed it's endpoint ({worker.endpoint[0]}:{worker.endpoint[1]} -> {addr[0]}:{addr[1]})")
                    worker.endpoint = addr

                worker.update_timestamp()

                conn.send(f"ACCEPT\n".encode("UTF-8"))

                remaining_data = int(header[4])
                conn.settimeout(CONNECTION_TIMEOUT_DURING_LOCK)
                with open(f"chunks/{header[2]}-{header[3]}.mkv", "wb") as f:
                    while remaining_data:
                        new_data = conn.recv(remaining_data)

                        if len(new_data) == 0:
                            connection_alive = False
                            conn.close()
                            break

                        f.write(new_data)
                        remaining_data -= len(new_data)

                conn.settimeout(SERVER_SOCKET_TIMEOUT)
                conn.send("ACCEPTED\n".encode("UTF-8"))

                open("ready.txt", "a").write(f"{header[2]}\n")
                chunk.done = True
        elif header[0] == "REUPLOAD":
            if file_hash != header[1]:
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] Hash mismatch: expected {file_hash}, got {header[1]}")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break
            
            if not header[2].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid seq_id: not an int (got {header[2]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[3].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid worker_id: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[4].isdigit():
                print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid chunk_size: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            with lock:
                chunk = chunk_queue.get_chunk_by_id(int(header[2]))

                if not chunk:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid seq_id: no chunk with id {header[2]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                worker = chunk.get_worker_by_id(int(header[3]))

                if not worker:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] invalid worker_id: no worker with id {header[3]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                if worker.endpoint != addr:
                    print(f"[{addr[0]}:{addr[1]} - UPLOAD] Warning: worker {worker.id} for chunk {chunk.seq_id} changed it's endpoint ({worker.endpoint[0]}:{worker.endpoint[1]} -> {addr[0]}:{addr[1]})")
                    worker.endpoint = addr

                worker.update_timestamp()

                if os.path.exists(f"chunks/{header[2]}-{header[3]}.mkv"):
                    current_size = os.stat(f"chunks/{header[2]}-{header[3]}.mkv").st_size
                else:
                    current_size = 0

                conn.send(f"RESUME {current_size}\n".encode("UTF-8"))

                remaining_data = int(header[4]) - current_size
                conn.settimeout(CONNECTION_TIMEOUT_DURING_LOCK)
                with open(f"chunks/{header[2]}-{header[3]}.mkv", "wb") as f:
                    while remaining_data:
                        new_data = conn.recv(remaining_data)

                        if len(new_data) == 0:
                            connection_alive = False
                            conn.close()
                            break

                        f.write(new_data)
                        remaining_data -= len(new_data)

                conn.settimeout(SERVER_SOCKET_TIMEOUT)
                conn.send("ACCEPTED\n".encode("UTF-8"))

                open("ready.txt", "a").write(f"{header[2]}\n")
                chunk.done = True
        elif header[0] == "ABORT":
            if file_hash != header[1]:
                print(f"[{addr[0]}:{addr[1]} - ABORT] Hash mismatch: expected {file_hash}, got {header[1]}")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break
            
            if not header[2].isdigit():
                print(f"[{addr[0]}:{addr[1]} - ABORT] invalid seq_id: not an int (got {header[2]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            if not header[3].isdigit():
                print(f"[{addr[0]}:{addr[1]} - ABORT] invalid worker_id: not an int (got {header[3]} instead)")
                conn.send("RESET\n".encode("UTF-8"))
                conn.close()
                connection_alive = False
                break

            with lock:
                chunk = chunk_queue.get_chunk_by_id(int(header[2]))

                if not chunk:
                    print(f"[{addr[0]}:{addr[1]} - ABORT] invalid seq_id: no chunk with id {header[2]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                worker = chunk.get_worker_by_id(int(header[3]))

                if not worker:
                    print(f"[{addr[0]}:{addr[1]} - ABORT] invalid worker_id: no worker with id {header[3]}")
                    conn.send("RESET\n".encode("UTF-8"))
                    conn.close()
                    connection_alive = False
                    break

                if worker.endpoint != addr:
                    print(f"[{addr[0]}:{addr[1]} - ABORT] Warning: worker {worker.id} for chunk {chunk.seq_id} changed it's endpoint ({worker.endpoint[0]}:{worker.endpoint[1]} -> {addr[0]}:{addr[1]})")
                    worker.endpoint = addr

                print(f"[{addr[0]}:{addr[1]} - ABORT] Removing worker {worker.id} from chunk {chunk.seq_id}, leaving chunk with {len(chunk.assigned_workers)} workers")
                chunk.remove_worker(worker)
                print(f"[{addr[0]}:{addr[1]} - ABORT] Updated chunk {chunk.seq_id}, it is now \"{chunk.status()}\"")
        else:
            conn.send("UNSUPPORTED\n".encode("UTF-8"))


print(f"[init:network] Listening on {sys.argv[4]}")

print(f"[init:cleanup] Starting cleanup job...")
cleanup_thread = threading.Thread(target = thread_cleanup_job, args = [])
cleanup_thread.start()
print(f"[init:cleanup] Cleanup job running")

print("[init:finish] Initialization complete")
while True:
    conn, addr = main_server_socket.accept()
    with lock:
        new_worker_thread = threading.Thread(target = connection_processor, args = [conn, addr])
        client_threads.append(new_worker_thread)
        new_worker_thread.start()
